Uses Crt,Mus,Graph,Msecs;

{
               PATHFINDING WITH DIRECT LINE MOVEMENT

  Does not necessary find the shortest path, but very often a good path.
  Better backtracking based on direct lines makes this method better
  looking and it usually finds better paths. The demo illustrates how
  it finds waypoints that the unit must move to (in a direct line) to
  avoid the obstacles. These waypoints reqire little memory storage
  and is therefore a good approach to pathfinding. Remember that you
  must be sure that your unit moves in the same line as the pathfinder
  marks for best result.

              John Christian Lonningdal  - 5. May 1996

                    Email: john@lis.pitt.edu

}


CONST
 ScrWidth       = 640;
 ScrHeight      = 470;
 TileSize       = 10;    { Modify this to make squares larger or smaller }
 MapWidth       = ScrWidth DIV TileSize;
 MapHeight      = ScrHeight DIV TileSize;
 MXTWidth       = (MapWidth-1)*TileSize;
 MXTHeight      = (MapHeight-1)*TileSize;

TYPE
 MapCoord       = RECORD
                   X,Y  : Word;
                  END;

VAR
 AI,BI,CI,
 XI,YI          : Integer;
 MX,MY          : Word;
 Tast           : Char;
 CollMap        : ARRAY[0..MapWidth-1,0..MapHeight-1] of Byte;
 LenMap         : ARRAY[0..MapWidth-1,0..MapHeight-1] of Word;
 StartPos,
 EndPos         : MapCoord;

{-------------------------------------------------------------------------}
Procedure GfxMode( M : Integer; Where : String);
VAR
 Driver,Mode : Integer;

 BEGIN
  Mode := M;
  Driver := Detect;
  InitGraph(Driver, Mode,Where);
  if GraphResult < 0 then
   BEGIN
    Writeln('ERROR: Graphics driver not installed!');
    Halt(1);
   END;
 END;


{-------------------------------------------------------------------------}
Procedure DrawGrid;
 BEGIN
  SetColor(1);
  For YI := 0 to MapHeight-1 do
   BEGIN
    Line(0,YI*TileSize,MXTWidth,YI*TileSize);
   END;

  For XI := 0 to MapWidth-1 do
   BEGIN
    Line(XI*TileSize,0,XI*TileSize,MXTHeight);
   END;
 END;

{-------------------------------------------------------------------------}
Procedure MakeBorder;
 BEGIN
  For YI := 0 to MapHeight-2 do
   BEGIN
    CollMap[0,YI] := 1;
    CollMap[MapWidth-2,YI] := 1;
   END;

  For XI := 0 to MapWidth-2 do
   BEGIN
    CollMap[XI,0] := 1;
    CollMap[XI,MapHeight-2] := 1;
   END;
 END;

{-------------------------------------------------------------------------}
Procedure DrawMap;
 BEGIN
  SetFillStyle(1,2);
  For YI := 0 to MapHeight-2 do
   For XI := 0 to MapWidth-2 do
    BEGIN
     If CollMap[XI,YI]>0 then
      Bar((XI*TileSize)+1,(YI*TileSize)+1,
          (XI*TileSize)+TileSize-1,(YI*TileSize)+TileSize-1);
    END;
 END;

{-------------------------------------------------------------------------}
Procedure DrawCollisionMap;
 BEGIN
  Setcolor(WHITE);
  OutTextxy(10,ScrHeight,'A - place start    B - place goal    SPACE - Find path');
  Tast:=#1;
  VisMus;
  MusOmraade(0,0,MXTWidth-1,MXTHeight-1);
  Repeat

   MX := MusX DIV TileSize;
   MY := MusY DIV TileSize;
   XI := MX*TileSize;
   YI := MY*TileSize;
   If MusKnapp=1 then
    BEGIN
     If NOT ((MX=StartPos.X) and (MY=StartPos.Y)) and
        NOT ((MX=EndPos.X)   and (MY=EndPos.Y)) then
      BEGIN
       GjemMus;
       SetFillStyle(1,2);
       Bar(XI+1,YI+1,XI+TileSize-1,YI+TileSize-1);
       VisMus;
       CollMap[MX,MY] := 1;
      END;
    END;
   If MusKnapp=2 then
    BEGIN
     If NOT ((MX=StartPos.X) and (MY=StartPos.Y)) and
        NOT ((MX=EndPos.X)   and (MY=EndPos.Y)) then
      BEGIN
       GjemMus;
       SetFillStyle(1,0);
       Bar(XI+1,YI+1,XI+TileSize-1,YI+TileSize-1);
       VisMus;
       CollMap[MX,MY] := 0;
      END;
    END;
   If Keypressed then
    BEGIN
     Tast := Readkey;
     Case Tast of
      'A','a' : BEGIN
                 If CollMap[MX,MY]=0 then
                  BEGIN
                   GjemMus;
                   SetFillStyle(1,0);
                   Bar((StartPos.X*TileSize)+1,(StartPos.Y*TileSize)+1,
                       (StartPos.X*TileSize)+TileSize-1,
                       (StartPos.Y*TileSize)+TileSize-1);
                   StartPos.X := MX; StartPos.Y := MY;
                   SetFillStyle(1,3);
                   Bar(XI+1,YI+1,XI+TileSize-1,YI+TileSize-1);
                   VisMus;
                  END;
                END;
      'B','b' : BEGIN
                 If CollMap[MX,MY]=0 then
                  BEGIN
                   GjemMus;
                   SetFillStyle(1,0);
                   Bar((EndPos.X*TileSize)+1,(EndPos.Y*TileSize)+1,
                       (EndPos.X*TileSize)+TileSize-1,
                       (EndPos.Y*TileSize)+TileSize-1);
                   EndPos.X := MX; EndPos.Y := MY;
                   SetFillStyle(1,4);
                   Bar(XI+1,YI+1,XI+TileSize-1,YI+TileSize-1);
                   VisMus;
                  END;
                END;
     End;
    END;
  Until (Tast=#27) or (Tast=' ');
  GjemMus;
 END;


{-------------------------------------------------------------------------}
Procedure FindTheWay;
VAR
 Pos         : MapCoord;
 NewPos      : MapCoord;
 Dir         : Byte;
 Step        : Byte;
 ShPos       : MapCoord;
 Plist       : ARRAY[0..500] of MapCoord; { List to remember all free paths }
 Llist       : ARRAY[0..500] of Word; { List to remember lenght of free paths}
 PCnt        : Word;
 GoalReached : Boolean;
 Txt         : String;
 Shortest    : Word;
 DiffX,DiffY : Integer;
 Loops       : Word;
 ShStep      : Word;
 StepCnt     : Word;
 AW          : Word;
 BestLine    : MapCoord;
 LineStartPos: MapCoord;
 Steps  : Word;

{-------------------------------------------------------------------------}
{ Rotate right to next direction to scan for free sqares.
  The rotation jumps two steps each time, except when going from
  left where it jumps to up-n-left to start scanning the across directions }
Procedure RotateRight;
 BEGIN
  Case Dir of
    3 : BEGIN Inc(NewPos.X,2);                  Dir:=6;  END;
    2 : BEGIN Inc(NewPos.X);   Inc(NewPos.Y);   Dir:=4;  END;
    6 : BEGIN                  Inc(NewPos.Y,2); Dir:=12; END;
    4 : BEGIN Dec(NewPos.X);   Inc(NewPos.Y);   Dir:=8;  END;
   12 : BEGIN Dec(NewPos.X,2);                  Dir:=9;  END;
    8 : BEGIN Dec(NewPos.X);   Dec(NewPos.Y);   Dir:=1;  END;
    9 : BEGIN Inc(NewPos.X);   Dec(NewPos.Y,2); Dir:=2;  END;
    1 : BEGIN                  Dec(NewPos.Y);   Dir:=3;  END;
  End;
 END;

  {---------------------------------------------------}
  Function LineIntersect(draw : boolean) : boolean;
  VAR
   i      : Word;
   x,y    : Real;
   Dx, Dy : Integer;
   Sx, Sy : Real;

   BEGIN
    Steps:=0;
    LineIntersect := False;

    AW:=(TileSize SHR 1);
    Dx := LineStartPos.X-Pos.X;
    Dy := LineStartPos.Y-Pos.Y;
    x := LineStartPos.X;
    y := LineStartPos.Y;
    If (Abs(Dx)>Abs(Dy)) then
     BEGIN
      if (Dx>0) then Sx := -1 else Sx := 1;
      If (Dx=0) then Sy := 0
                else Sy := Dy/(Dx*Sx);
      Steps := Abs(dx);
     END ELSE
     BEGIN
      if (Dy>0) then Sy := -1 else Sy := 1;
      If (Dy=0) then Sx := 0
                else Sx := Dx/(Dy*Sy);
      Steps := Abs(dy);
     END;

    For i:=0 to Steps do
     BEGIN
      If (CollMap[Round(X),Round(Y)]>0) then LineIntersect := True;
      If ((draw) AND (i>0)) then
       BEGIN
        SetFillStyle(1,6);
        Bar((Round(X)*TileSize)+1,(Round(Y)*TileSize)+1,
            (Round(X)*TileSize)+TileSize-1,(Round(Y)*TileSize)+TileSize-1);

       END;
      y := y + Sy;
      x := x + Sx;
     END

   END;


  {---------------------------------------------------}
  Procedure DrawArrows;
  VAR
   AW : Word;

   BEGIN
     AW:=(TileSize SHR 1);
       Line((Pos.X*TileSize)+AW,(Pos.Y*TileSize)+AW,
            (NewPos.X*TileSize)+AW,(NewPos.Y*TileSize)+AW);
   END;


 BEGIN

  SetFillStyle(1,0);
  Bar(0,ScrHeight,500,GetmaxY);

  Tast:=#1;
  PCnt := 1;
  PList[0] := StartPos;
  DiffX := (StartPos.X-EndPos.X);
  DiffY := (StartPos.Y-EndPos.Y);
{  LList[0] := Abs(DiffX)+Abs(DiffY);}
  LList[0] := (DiffX*DiffX)+(DIffY*DiffY);

  LenMap[StartPos.X,StartPos.Y]:=1;
  Pos := StartPos;
  NewPos := Pos;
  Step := 1;
  GoalReached := False;
  SetColor(8);
  MSStart_Clock;
  Repeat

     { Find the one with the shortest distance in the list }

     Shortest:=0;
     For CI := 0 to PCnt-1 do
      BEGIN
       If LList[CI]<LList[Shortest] then
        BEGIN
         Shortest:=CI;
        END;
      END;
     { Get the one with the shortest distance }
     Pos := PList[Shortest];
     { Replace this one with the last one in the list }
     PList[Shortest] := PList[PCnt-1];
     LList[Shortest] := LList[PCnt-1];
     { Make the list one point smaller }
     Dec(PCnt);

     NewPos := Pos;
     Dir := 2; Dec(NewPos.Y); { Start with direction upwards }
     For BI := 0 to 7 do { Scan all 8 directions to find free ways to go }
      BEGIN
       {Check to see if there is a free square here}
       If ((LenMap[NewPos.X,NewPos.Y]=65535) AND
          (CollMap[NewPos.X,NewPos.Y]=0)) then
        BEGIN
         { Record stepvalue to get into this square }
         StepCnt := LenMap[Pos.X,Pos.Y]+1;
         LenMap[NewPos.X,NewPos.Y] := StepCnt;
         PList[PCnt] := NewPos;
         { Add the distance to list also }
         DiffX := (NewPos.X-EndPos.X);
         DiffY := (NewPos.Y-EndPos.Y);
{         LList[PCnt] := Abs(DiffX)+Abs(DiffY)+StepCnt;}
         LList[PCnt] := (DiffX*DiffX)+(DIffY*DiffY);
         Inc(PCnt);
         DrawArrows; {Remove for faster view of only shortest path}
        END;

       {End the checking if we have reached the endpoint}
       If (NewPos.X=EndPos.X) and (NewPos.Y=EndPos.Y) then
        BEGIN
         GoalReached := True;
         BI:=7;
        END;
        RotateRight;
      END;

  Until (GoalReached) or (PCnt=0);

  MSStop_Clock;

  Str(MSTimer,Txt);
  SetColor(15);
  OutTextxy(290,ScrHeight,Txt);

  { backtrack from B to A to find the path, always trying to find the
    shortest stepvalue back }
  SetColor(15);

  Step:=1;
  If GoalReached then
   BEGIN
    Pos := EndPos;
    LineStartPos := Pos;

    While NOT ((Pos.X=StartPos.X) AND (Pos.Y=StartPos.Y)) do
     BEGIN
{      Inc(Step);}
      NewPos := Pos;
      ShStep := 65535;
      Dir := 2; Dec(NewPos.Y); { Start with direction upwards }
      For AI:=0 to 7 do
       BEGIN
        If LenMap[NewPos.X,NewPos.Y]<ShStep then
         BEGIN
          ShStep:=LenMap[NewPos.X,NewPos.Y];
          ShPos := NewPos;
         END;
        RotateRight;
       END;

      Pos := ShPos;

      { Scan for the best path of line that has no intersections }
      if (NOT LineIntersect(FALSE)) then BestLine := Pos;

      if ((Pos.X=StartPos.X) AND (Pos.Y=StartPos.Y))  then
       BEGIN
        Pos := BestLine;
        LineIntersect(TRUE);
        Inc(Step,Steps);
        SetFillStyle(1,14);
        Bar((Pos.X*TileSize)+1,(Pos.Y*TileSize)+1,
            (Pos.X*TileSize)+TileSize-1,(Pos.Y*TileSize)+TileSize-1);
        AW:=(TileSize SHR 1);
        Line((LineStartPos.X*TileSize)+AW,(LineStartPos.Y*TileSize)+AW,
             (Pos.X*TileSize)+AW,(Pos.Y*TileSize)+AW);
        LineStartPos := Pos;
       END;

     END;

    Str(Step-1,Txt);
    OutTextxy(10,ScrHeight,'The goal was reached in '+Txt+' steps.');
   END ELSE
   BEGIN
    OutTextxy(10,ScrHeight,'No way to reach the goal! Ugh!');
   END;
 END;


{-------------------------------------------------------------------------}
Begin
 GfxMode(0,'');

 For YI := 0 to MapHeight-1 do
  For XI := 0 to MapWidth-1 do
   CollMap[XI,YI] := 0;

 MakeBorder;

 StartPos.X := 5; StartPos.Y := 5;
 EndPos.X := 10; EndPos.Y := 10;

 Repeat

  For YI := 0 to MapHeight-1 do
   For XI := 0 to MapWidth-1 do
     LenMap[XI,YI] := 65535;

  ClearDevice;
  DrawGrid;
  DrawMap;
  SetFillStyle(1,4);
  Bar((EndPos.X*TileSize)+1,(EndPos.Y*TileSize)+1,
      (EndPos.X*TileSize)+TileSize-1,
      (EndPos.Y*TileSize)+TileSize-1);
  SetFillStyle(1,3);
  Bar((StartPos.X*TileSize)+1,(StartPos.Y*TileSize)+1,
      (StartPos.X*TileSize)+TileSize-1,
      (StartPos.Y*TileSize)+TileSize-1);

  Setcolor(WHITE);
  OutTextxy(550,ScrHeight,'ESC - quit');
  DrawCollisionMap;
  If (Tast<>#27) then
   BEGIN
    FindTheWay;
    Tast := Readkey;
   END;

 Until (Tast=#27);

 CloseGraph;
End.