/* $Id$
 *
 * History:101,52
 *
 * $Author$
 * $Date$
 * $Revision$
 *
 * Copyright (C) 1995 Pragmatic Designs, All Rights Reserved
 *
 * $Log$
 *
 */


#include <mem.h>
#include <stdio.h>

#include <fastgraf.h>
#include <fgfonts.h>

#include "struct.h"
#include "prototyp.h"


/****************************************************************
 * Calculate the level of influence for an even column.
 */
int
calcEvnMap (const iPoint* loc)
{
	int		rv;							// return value
	int		x, y;

	x = loc->x;
	y = loc->y;
	rv = oldMap[x][y];

	y--;								// upper hex
	if (y >= 0) {
		if (oldMap[x][y] != 0) {
			rv += (oldMap[x][y] > 0) ? 1 : -1;
		}
	}

	x++;								// upper right hex
	if ((x < 23) && (y >= 0)) {
		if (oldMap[x][y] != 0) {
			rv += (oldMap[x][y] > 0) ? 1 : -1;
		}
	}

	y++;								// lower right hex
	if (x < 23) {
		if (oldMap[x][y] != 0) {
			rv += (oldMap[x][y] > 0) ? 1 : -1;
		}
	}

	x--;								// lower hex
	y++;
	if (y < 23) {
		if (oldMap[x][y] != 0) {
			rv += (oldMap[x][y] > 0) ? 1 : -1;
		}
	}

	x--;								// lower left hex
	y--;
	if (x >= 0) {
		if (oldMap[x][y] != 0) {
			rv += (oldMap[x][y] > 0) ? 1 : -1;
		}
	}

	y--;								// upper left hex
	if ((x >=0) && (y >= 0)) {
		if (oldMap[x][y] != 0) {
			rv += (oldMap[x][y] > 0) ? 1 : -1;
		}
	}

	return (rv);
}


/****************************************************************
 * Calculate the level of influence for an odd column.
 */
int
calcOddMap (const iPoint* loc)
{
	int		rv;							// return value
	int		x, y;

	x = loc->x;
	y = loc->y;

	rv = oldMap[x][y];

	if (y == 22) return (rv);			// dead hex

	y--;								// upper hex
	if (y >= 0) {
		if (oldMap[x][y] != 0) {
			rv += (oldMap[x][y] > 0) ? 1 : -1;
		}
	}

	x++;								// upper right hex
	y++;
	if (x < 23) {
		if (oldMap[x][y] != 0) {
			rv += (oldMap[x][y] > 0) ? 1 : -1;
		}
	}

	y++;								// lower right hex
	if ((x < 23) && ( y < 23)) {
		if (oldMap[x][y] != 0) {
			rv += (oldMap[x][y] > 0) ? 1 : -1;
		}
	}

	x--;								// lower hex
	if (y < 23) {
		if (oldMap[x][y] != 0) {
			rv += (oldMap[x][y] > 0) ? 1 : -1;
		}
	}

	x--;								// lower left hex
	if (x >= 0) {
		if (oldMap[x][y] != 0) {
			rv += (oldMap[x][y] > 0) ? 1 : -1;
		}
	}

	y--;								// upper left hex
	if (x >= 0) {
		if (oldMap[x][y] != 0) {
			rv += (oldMap[x][y] > 0) ? 1 : -1;
		}
	}

	return (rv);
}


/****************************************************************
 * Calculate the influence map.
 */
void
calcInfluence (int count)
{
	int		i;							// counter
	iPoint	loc;						// counters

	for (loc.y = 0; loc.y < SCRN_W; loc.y++) {
		for (loc.x = 0; loc.x < SCRN_H; loc.x++) {
			oldMap[loc.x][loc.y] = tstMap[loc.x][loc.y] * count;
		}
	}

	for (i = 0; i < count; i++) {
		for (loc.y = 0; loc.y < SCRN_W; loc.y++) {
			for (loc.x = 0; loc.x < SCRN_H; loc.x++) {
				if (loc.x %2) {
					newMap[loc.x][loc.y] = calcOddMap (&loc);
				} else {
					newMap[loc.x][loc.y] = calcEvnMap (&loc);
				}
			}
		}
		/* Copy newMap to oldMap.
		 */
		memcpy (&oldMap, &newMap, sizeof (oldMap));
	}
}


/****************************************************************
 * Show current level of influence
 */
void
showInfluence (void)
{
	int		i;							// counter
	char	s[5];

	ScrnMap*	pMap   = scrnMap;
	int*		pInflu = &oldMap[0][0];

	for (i = 0; i < SCRN_H * SCRN_W; i++) {
		if (pMap->visible) {
			fg_setcolor (BLACK);		// erase the old value
			fg_rect (pMap->vertex.x - 9, pMap->vertex.x + 9,
					 pMap->vertex.y + 7, pMap->vertex.y + 13);

			fg_move (pMap->vertex.x, pMap->vertex.y + 12);

			fg_setcolor (D_GRAY);		// default color
			if (*pInflu > 0) fg_setcolor (L_BLUE);
			if (*pInflu < 0) fg_setcolor (RED);

			sprintf (s, "%d", *pInflu);
			fgf_print (s, 4);
		}
		pMap++;
		pInflu++;
	}
}

/* EOF - $RCSfile$ */
