/* $Id$
 *
 * History:270,33
 *
 * $Author$
 * $Date$
 * $Revision$
 *
 * Copyright (C) 1995 Pragmatic Designs, All Rights Reserved
 *
 * $Log$
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <alloc.h>
#include <string.h>

#include <fastgraf.h>
#include <fgfonts.h>

#include "struct.h"
#include "prototyp.h"

/****************************************************************
 * The base hex, as defined for Fastgraph.
 */
Hex	BaseHex = {{
	{ -10, -16 },
	{ -20,   0 },
	{ -10,  16 },
	{  10,  16 },
	{  20,   0 },
	{  10, -16 }
}};


/****************************************************************
 * Local prototypes.
 */
int		_pointInHex (const iPoint* mouse, int ptrOff);


/****************************************************************
 * Inz the screen map table.
 */
void
initScreenMap (void)
{
	size_t	size;
	int		col, row;					// column & row counters
	int		gcx, gcy;					// calculated center point values

	ScrnMap*	pMap;					// start of the map table

	int		hexVis;						// will hex be visible flag?

	/* Alloc the memory for the screen map table.
	 */
	size = (SCRN_W * SCRN_H) * sizeof (ScrnMap);
	scrnMap = (ScrnMap*) malloc (size);
	if (scrnMap == NULL) {				// didn't take
		Error ("Unable to malloc memory for the screen map table.\n");
	}

	pMap = scrnMap;						// inz the pointer
	gcx = 20;							// center of the first column

	/* For each column...
	 */
	for (col = 0; col < SCRN_W; col++) {
		gcy = ((gcx /2) %2) ? 32 : 16;	// odd / even starting value

		/* For each row...
		 */
		for (row = 0; row < SCRN_H; row++) {
			pMap->vertex.x = gcx;		// save the center point values
			pMap->vertex.y = gcy;

			/* Check for visibility.  0 is not visible, 1 is.
			 * Note the inverted logic at the assignment line.
			 */
			hexVis = (col % 2) && (row == SCRN_H - 1);
			pMap->visible = !hexVis;

			pMap->loc.x = col +1;		// save the coordinate data
			pMap->loc.y = row +1;

			pMap++;						// next map location
			gcy += 32;					// next row center point
		}
		gcx += 30;						// next column center point
	}									// (3/4 of total width)
}


/****************************************************************
 * Draw the hex grid.
 */
void
drawGrid (void)
{
	int		i;							// counter

	ScrnMap*	pMap = scrnMap;	// start of the screen map

	/* For all hexes...
	 */
	for (i = 0; i < (SCRN_W * SCRN_H); i++) {
		if (pMap->visible) {			// if hex is visible...
			fg_polyoff (pMap->vertex.x, pMap->vertex.y);
			fg_polyline ((int*) BaseHex.vertex, 6);
//			fg_point (pMap->vertex.x, pMap->vertex.y);
		}
		pMap++;							// next hex
	}
	fg_polyoff (0, 0);					// reset polyoff
}


/****************************************************************
 * Test to see which hex the mouse was clicked in.  This function
 * converts the passed mouse click values, in screen coordinates,
 * and returns the XY offset values in "result."  A flag is returned
 * indicating a valid hex location was found.
 */
int
pointInHex (iPoint* result, const iPoint* mouse)
{
	iPoint	click;						// working storage
	int		pHex;						// offset counter

	int		rv;							// gp return value

	/* Initialize the offset pointer.
	 */
	click.x = mouse->x;					// get the passed screen location
	click.y = mouse->y;

	click.x /= 30;						// find the column, (3/4 width)

	if (click.x % 2) {					// even or odd column?
		click.y -= 16;					// odd column, adjust
	}

	click.y /= 32;						// full height

	/* Now click.x and click.y should equal the number of columns
	 * across and the number of rows down, respectfully.  Now we
	 * see if this is true...
	 */
	pHex = (SCRN_H * click.x) + click.y;

	/* See if this is the correct hex.
	 */
	rv = _pointInHex (mouse, pHex);
	if (rv) {
		result->x = click.x;			// save the hex coordinates
		result->y = click.y;
		return 1;
	}

	/* It's not that hex, pHex is one column off to the right.
	 * Move it over and look at the hex to the lower right.
	 * The adjustment value depends on whether the column
	 * value is odd or even.
	 */
	pHex -= (click.x % 2) ? SCRN_H -1 : SCRN_H;

	rv = _pointInHex (mouse, pHex);
	if (rv) {
		result->x = click.x -1;			// save the hex coordinates
		result->y = (click.x % 2) ? click.y +1 : click.y;
		return 1;
	}

	/* Not that hex, must be the upper right hex.
	 */
	pHex--;
	rv = _pointInHex (mouse, pHex);
	if (rv) {
		result->x = click.x -1;			// save the hex coordinates
		result->y = (click.x % 2) ? click.y : click.y -1;
		return 1;
	}
	/* None of the three, must be a partial hex around the
	 * edge of the screen.
	 */
	return 0;							// fail
}


/****************************************************************
 * Test for a point inside a hexagon.  Enter with the XY coordinates
 * and the offset into the poly center array.
 */
int
_pointInHex (const iPoint* mouse, int ptrOff)
{
	int		rv = 0;						// gp return value
	ScrnMap* pMap = scrnMap;			// start of the offset table

	pMap += ptrOff;						// adjust the pointer

	if (pMap->visible) {				// if hex is visible...
		/* Test for a hit inside the hexagon.
		 */
		fg_polyoff (pMap->vertex.x, pMap->vertex.y);
		rv = (fg_inside ((int*) BaseHex.vertex, 6, mouse->x, mouse->y));
		fg_polyoff (0, 0);
	}
	return rv;
}


/****************************************************************
 * Free the screen memory.
 */
void
freeScreenMap (void)
{
	if (scrnMap != NULL) {
		free ((void*) scrnMap);
		scrnMap = NULL;
	}
}


/****************************************************************
 * Label the grid.
 */
void
labelGrid (void)
{
	int		i;							// counter
	char	s[5];

	ScrnMap*	pMap = scrnMap;

	for (i = 0; i < (SCRN_W * SCRN_H); i++) {
		if (pMap->visible) {
			fg_move (pMap->vertex.x, pMap->vertex.y -9);
			sprintf (s, "%02d%02d", pMap->loc.x, pMap->loc.y);
			fgf_print (s, 4);
		}
		pMap++;
	}
}


/****************************************************************
 * (DEBUG) Outline the selected hex.
 * Select color based on underlaying color.
 */
void
outlineHex (const iPoint* loc)
{
	ScrnMap* pMap = scrnMap;
	int*	pInflu = &tstMap[0][0];

	pMap += (SCRN_H * loc->x) + loc->y;
	pInflu += (SCRN_H * loc->x) + loc->y;

	if (*pInflu ==  1) fg_setcolor (L_BLUE);
	if (*pInflu == -1) fg_setcolor (RED);
	if (*pInflu ==  0) fg_setcolor (M_GRAY);

	if (pMap->visible) {
		fg_polyoff (pMap->vertex.x, pMap->vertex.y);
		fg_polyline ((int*) BaseHex.vertex, 6);
//		fg_point (pMap->vertex.x, pMap->vertex.y);
	}
}


/****************************************************************
 * Scroll the window with the mouse.
 */
void
testPos (const iPoint* loc)
{
	static int	hpo, vpo = 0;

	int		update = 0;

	if (loc->x < hpo) {
		hpo = loc->x;
		update = 1;
	} else if (loc->x > hpo + 640) {
		hpo = loc->x - 640 + 1;
		update = 1;
	} else if (loc->y < vpo) {
		vpo = loc->y;
		update = 1;
	} else if (loc->y > vpo + 480) {
		vpo = loc->y - 480 + 1;
		update = 1;
	}

	if (update) {
		fg_pan (hpo, vpo);
	}
}

/* EOF - $RCSfile$ */
