//	World engine
// (C) 1997 Aurel Balmosan

#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <math.h>

#include "World.h"


LevelMap::LevelMap(int w, int h)
{       value=NULL; ptr=NULL; map=NULL;
        width=height=0;
        if (w>0 && h>0)
        {       width=w; height=h;
                value=(int*)malloc(sizeof(int)*h);
                ptr=(int**)malloc(sizeof(int*)*h);
                map=(int*)malloc(sizeof(int)*w*h);
                for (int i=0; i<h; i++) ptr[i]=&map[i*w];
                memset(value, 0, sizeof(int)*height);
                memset(map, 0, sizeof(int)*height*width);
        }
}

void LevelMap::SetSize(int w, int h)
{       if (value) { delete[] value; value=NULL; }
        if (ptr) { delete [] ptr; ptr=NULL; }
        if (map) { delete [] map; map=NULL; }
        width=height=0;
        if (w>0 && h>0)
        {       width=w; height=h;
                value=(int*)malloc(sizeof(int)*h);
                ptr=(int**)malloc(sizeof(int*)*h);
                map=(int*)malloc(sizeof(int)*w*h);
                for (int i=0; i<h; i++) ptr[i]=&map[i*w];
                memset(value, 0, sizeof(int)*h);
                memset(map, 0, sizeof(int)*h*w);
        }
}


LevelMap::~LevelMap()
{       if (value) { free(value); value=NULL; }
        if (ptr) { free(ptr); ptr=NULL; }
        if (map) { free(map); map=NULL; }
        width=height=0;
}

void LevelMap::Generate(int iterations, int seed)
{       Generate(iterations, Point(0,0), Point(width,height), seed);
}

void LevelMap::Generate(int iterations, Point min, Point max, int seed)
{       double  pfi, cp, sp, dl;
        int     i, h, j, x, y, l, dd;
        int     vmax, vmin, px, py, wmod, hmod;

        if (width<=0 || height<=0) return;

        if (min.x<0) min.x=0;
        if (min.y<0) min.y=0;
        if (max.x>width) max.x=width;
        if (max.y>height) max.x=height;
        wmod=max.x-min.x; hmod=max.y-min.y;

        if (seed) rnd.Seed(seed);
        memset(value, 0, sizeof(int)*height);
        memset(map, 0, sizeof(int)*height*width);
        while (iterations-->0)
        {       px=rnd.Value(wmod)+min.x; py=rnd.Value(hmod)+min.y;
                pfi=(double)rnd.Value(1000001)*2.0*PI/1000001.0;
                cp=cos(pfi); sp=sin(pfi);
                h=(pfi>PI/2 && pfi<=PI+PI/2) ? -1 : 1;
                if (sp>-1e-20 && sp<1e-20)
                {       for (y=0; y<py; y++) value[y]+=h;
                        for (   ; y<height; y++) value[y]-=h;
                }
                else
                {       dl=cp/sp;
                        if (dl>fix_max/2) dl=fix_max/2;
                        else if (dl<-fix_max/2) dl=-fix_max/2;
                        dd=(int)(fix1*dl);
                        for (l=dd,y=py+1,x=px+(l>>fix_shift); 
                                y<height && x>=0 && x<width; 
                                y++,l+=dd, x=px+(l>>fix_shift))
                        {       value[y]+=h;
                                ptr[y][x]-=h<<1;
                        }
                        if      (x<0)      for (;y<height;y++) value[y]-=h;
                        else if (x>=width) for (;y<height;y++) value[y]+=h;

                        for (l=0,y=py,x=px+(l>>fix_shift); 
                                y>=0 && x>=0 && x<width; 
                                y--,l-=dd, x=px+(l>>fix_shift))
                        {       value[y]+=h;
                                ptr[y][x]-=h<<1;
                        }
                        if      (x<0)      for (;y>=0;y--) value[y]-=h;
                        else if (x>=width) for (;y>=0;y--) value[y]+=h;
                }
        }

        vmin=vmax=map[0];
        for (y=0; y<height; y++)
        {       h=value[y];
                for (x=0; x<width; x++)
                {       h+=ptr[y][x]; 
                        ptr[y][x]=h;
                        if (h>vmax) vmax=h;
                        if (h<vmin) vmin=h;
                }
        }
        j=vmax-vmin; h=width*height;
        for (i=0; i<h; i++)
                map[i]=(map[i]-vmin)*fix1/j;
}
                

void LevelMap::Combine(const LevelMap *a, MapCombineFunc f)
{       int     wh=width*height;
        int     i, vmin, vmax, j;

        if (a->width!=width || a->height!=height) return;

        vmin=1<<30; vmax=-1<<30;
        for (i=0; i<wh; i++)
        {       map[i]=(int)(fix1*f((float)map[i]/fix1, (float)a->map[i]/fix1));
                if (map[i]>vmax) vmax=map[i];
                if (map[i]<vmin) vmin=map[i];
        }
        j=vmax-vmin;
        for (i=0; i<wh; i++)
                map[i]=(map[i]-vmin)*fix1/j;
}

void LevelMap::Combine(MapCombineFunc f)
{       int     wh=width*height;
        int     i, vmin, vmax, j;

        vmin=1<<30; vmax=-1<<30;
        for (i=0; i<wh; i++)
        {       map[i]=(int)(fix1*f((float)map[i]/fix1, 1.0));
                if (map[i]>vmax) vmax=map[i];
                if (map[i]<vmin) vmin=map[i];
        }
        j=vmax-vmin;
        for (i=0; i<wh; i++)
                map[i]=(map[i]-vmin)*fix1/j;
}
