
"""REX/Python

Based on Robert D. Cameron's REX/Perl 1.0.

Original copyright notice follows:

REX/Perl 1.0

Robert D. Cameron "REX: XML Shallow Parsing with Regular Expressions",
Technical Report TR 1998-17, School of Computing Science, Simon Fraser
University, November, 1998.
Copyright (c) 1998, Robert D. Cameron.
The following code may be freely used and distributed provided that
this copyright and citation notice remains intact and that modifications
or additions are clearly identified.

"""

import re

# An XML document
TextSE = "[^<]+"
UntilHyphen = "[^-]*-"
Until2Hyphens = UntilHyphen + "(?:[^-]" + UntilHyphen + ")*-"
CommentCE = Until2Hyphens + ">?"
UntilRSBs = "[^\\]]*](?:[^\\]]+])*]+"
CDATA_CE = UntilRSBs + "(?:[^\\]>]" + UntilRSBs + ")*>"
S = "[ \\n\\t\\r]+"
NameStrt = "[A-Za-z_:]|[^\\x00-\\x7F]"
NameChar = "[A-Za-z0-9_:.-]|[^\\x00-\\x7F]"
Name = "(?:" + NameStrt + ")(?:" + NameChar + ")*"
QuoteSE = "\"[^\"]*\"|'[^']*'"
DT_IdentSE = S + Name + "(?:" + S + "(?:" + Name + "|" + QuoteSE + "))*"
MarkupDeclCE = "(?:[^\\]\"'><]+|" + QuoteSE + ")*>"
S1 = "[\\n\\r\\t ]"
UntilQMs = "[^?]*\\?+"
PI_Tail = "\\?>|" + S1 + UntilQMs + "(?:[^>?]" + UntilQMs + ")*>"
DT_ItemSE = "<(?:!(?:--" + Until2Hyphens + ">|[^-]" + MarkupDeclCE + ")|\\?" + Name + "(?:" + PI_Tail + "))|%" + Name + ";|" + S
DocTypeCE = DT_IdentSE + "(?:" + S + ")?(?:\\[(?:" + DT_ItemSE + ")*](?:" + S + ")?)?>?"
DeclCE = "--(?:" + CommentCE + ")?|\\[CDATA\\[(?:" + CDATA_CE + ")?|DOCTYPE(?:" + DocTypeCE + ")?"
PI_CE = Name + "(?:" + PI_Tail + ")?"
EndTagCE = Name + "(?:" + S + ")?>?"
#AttValSE = "\"[^<\"]*\"|'[^<']*'|\\w+"  # STRICT
AttValSE = "\"[^<\"]*\"|'[^<']*'|[^\"' =>]+"
ElemTagCE = Name + "(?:" + S + Name + "(?:" + S + ")?=?(?:" + S + ")?(?:" + AttValSE + "))*(?:" + S + ")?/?>?"
MarkupSPE = "<(?:!(?:" + DeclCE + ")?|\\?(?:" + PI_CE + ")?|/(?:" + EndTagCE + ")?|(?:" + ElemTagCE + ")?)"
XML_SPE = TextSE + "|" + MarkupSPE

# Attributes within a tag
OS = r"[\s\n\r]*"
NameValuePair=OS+'(?P<NAME>'+Name+')'+OS+'(?:=?'+OS+'(?P<VALUE>'+AttValSE+'))'

# The original Perl subroutine
#
# sub ShallowParse {
#   my($XML_document) = @_;
#   return $XML_document =~ /$XML_SPE/g;
# }

def ShallowParse(XML_document, regex=XML_SPE):
        parts = []
        start = 0
        R = re.compile(regex)
        while start < len(XML_document):
            m = R.match(XML_document, start)
            if not m:
                parts.append(XML_document[start:])
                break
            parts.append(XML_document[m.start():m.end()])
            start = m.end()

        return parts
    
def _test():
        print ShallowParse('<MYTAG x="5" y="9">This is <I>shallow parsing</I>.</MYTAG>')

if __name__ == '__main__':
        _test()

# _test() output looks like
# ['<MYTAG>', 'This is ', '<I>', 'shallow parsing', '</I>', '.', = '</MYTAG>']
