from string import *
import time
import proxy3_options
from proxy3_util import *
from gtk import *

class Toggle:
    def __init__(self, box, label, option_name, tooltip):
        self.option_name = option_name
        self.button = new(GtkCheckButton, label=label,
                          parent=box, visible=TRUE)
        tt.set_tip(self.button, tooltip, '')
        self.button.connect("clicked", self.toggle)
        self.configure_button()
        
    def configure_button(self):
        self.button.set_active(proxy3_options.options.get(self.option_name))

    def toggle(self, *args):
        proxy3_options.options[self.option_name] = self.button.get_active()
        self.configure_button()
    
class TextOption:
    def __init__(self, box, label, option_name, tooltip):
        self.option_name = option_name
        self.field = new(GtkEntry, 
                          parent=box, visible=TRUE)
        tt.set_tip(self.field, tooltip, '')
        self.field.set_text(proxy3_options.options.get(option_name, ''))
        #self.button.connect("clicked", self.toggle)
        #self.configure_button()
        
    def configure_button(self):
        self.button.set_active(proxy3_options.options.get(self.option_name))

    def toggle(self, *args):
        proxy3_options.options[self.option_name] = self.button.get_active()
        self.configure_button()
    
def destroy(*args):
    window.hide()

def setup():
    global window, box, tt

    rc_parse("proxy.gtkrc")    
    time.sleep(1)
    
    tt = GtkTooltips()
    tt.set_delay(500)

    window = new(GtkWindow, type=WINDOW_TOPLEVEL, title="Proxy Configuration",
                 allow_grow=TRUE, allow_shrink=TRUE, border_width=5)
    window.set_default_size(400, 300)
    window.connect("destroy", destroy)

    container = new(GtkHPaned, parent=window)
    
    box = new(GtkScrolledWindow, visible=TRUE)
    box.set_usize(200, -1)
    container.add1(box)
    box.set_policy(POLICY_AUTOMATIC, POLICY_AUTOMATIC)
    box2 = new(GtkVBox, visible=TRUE)
    box2.set_homogeneous(1)
    box.add_with_viewport(box2)
    box = box2

    Toggle(box, "Filter ads", 'remove-ads',
           'Ad filtering removes ads from the HTML')
    Toggle(box, "Block ads", 'block-ads',
           'Ad blocking severs connections to ad sites')
    Toggle(box, "Blue Slashdot", 'blue-slashdot',
           'Change Slashdot from green on white to blue on gray')
    Toggle(box, "Remove Slashdot Ads", 'remove-slashdot-ads',
           'Work extra hard to remove Slashdot advertising')
    Toggle(box, "Remove NewsTracker Ads", 'remove-newstracker-ads',
           'Work extra hard to remove NewsTracker advertising')
    Toggle(box, 'Fix Microsoft Charset', 'fix-microsoft-charset',
           'Translate Microsoft characters into standard ASCII')
    Toggle(box, "Rewrite click-through URLs", 'rewrite-urls',
           'Rewrite forwarding URLs so that servers can\'t track your clicks')
    Toggle(box, 'Remove popups', 'remove-popups',
           'Remove a limited set of popup windows')
    Toggle(box, "Remove plugins", 'remove-plugins',
           'Remove EMBED/OBJECT tags (plugins) from pages')
    Toggle(box, "Remove JavaScript", 'remove-script',
           'Remove JavaScript from pages')
    #TextOption(box, 'Referer:', 'referer', 'Send this referer')

    box = new(GtkScrolledWindow, visible=TRUE)
    container.add2(box)
    box.set_policy(POLICY_AUTOMATIC, POLICY_AUTOMATIC)
    box2 = new(GtkVBox, visible=TRUE)
    box2.set_homogeneous(1)
    box.add_with_viewport(box2)
    box = box2

    w = new(GtkLabel, label="Connections:", parent=box)
    box.set_child_packing(w, 0, 0, 0, PACK_START)
    w.show()
    w = new(GtkHSeparator, parent=box)
    box.set_child_packing(w, 0, 0, 0, PACK_START)
    w.show()
    w = new(GtkVBox, parent=box)
    box.set_child_packing(w, 1, 1, 0, PACK_START)
    w.show()

    container.show()
    
    tt.enable()

    window.show()
    while 1:
        if not events_pending(): time.sleep(0.05)
        try:
            mainiteration(block=FALSE)
        except:
            import traceback
            traceback.print_exc(100, open('/tmp/proxy.traceback', 'a'))
            
def connection_to_key(conn): # Copied from mod_curses
    try: return conn.reqno
    except: return conn
    
ROWS = {}
def event_idle(connections):
    event_status([], connections, None)
    
def event_status(ready_list, connections, conn):
    global YSIZE
    # Check for stale connections
    if connections is not None:
        known_keys = map(connection_to_key, connections)
        for c in ROWS.keys():
            if c not in known_keys and time.time() - ROWS[c][0] > 10: #10 sec
                box.remove(ROWS[c][1])
                del ROWS[c]
                
    if not conn: return
    
    key = connection_to_key(conn)
    if not ROWS.has_key(key):
        w = new(GtkProgressBar, parent=box)
        box.set_child_packing(w, 0, 0, 0, PACK_START)
        w.set_usize(200, 24)
        w.set_text_alignment(0, 0.5)
        w.set_show_text(1)
        w.set_format_string(conn.__repr__(40))
        w.configure(0, 0, 100)
        w.show()
        ROWS[key] = [time.time(), w]

    w = ROWS[key][1]
    ROWS[key][0] = time.time()
    try: 
        url = conn.site+conn.document
        if conn.site=='www.google.com' and conn.document[:8]=='/search?':
            import cgi
            url = 'Search: %s' % cgi.parse_qs(conn.document[8:]).get('q', ['(no keywords)'])[0]
        else:
            url = shorten_url(conn.site+conn.document, 120)
        if conn.length:
            w.set_format_string('(%v%%) ' + url)
            percentage = int(1.0*conn.bytes/conn.length)
            if percentage < 0: percentage = 0
            if percentage > 100: percentage = 100
            w.set_percentage(percentage)
        else:
            w.set_format_string('%s bytes: %s' % (conn.bytes, url))
    except AttributeError:
        w.set_format_string(replace(conn.__repr__(40), '%', '%%'))
        
if __name__=='__main__':
    setup()
else:
    import thread
    thread.start_new_thread(setup, ())
