from string import *
import urlparse
import re
import proxy3_options, proxy3_rewrite
from proxy3_filter import *
from proxy3_event import broadcast_event

class NoScript(BufferSomeFilter):
    BUFFER_LEN = 8192
    
    """Search for scripts and remove them"""
    def filter(self, s):
        if not proxy3_options.options.get('remove-script', 0): return s
        s = remove_pattern(s, '<SCRIPT', '</SCRIPT>')
        s = remove_pattern(s, '<script', '</script>')
        return s

class NoScriptForce(BufferSomeFilter):
    "This is just like NoScript, but it ignores the configuration file option"
    BUFFER_LEN = 8192
    
    """Search for scripts and remove them, regardless of preference"""
    def filter(self, s):
        s = remove_pattern(s, '<SCRIPT', '</SCRIPT>')
        s = remove_pattern(s, '<script', '</script>')
        return s

class NoPlugins(BufferSomeFilter):
    """Search for scripts and remove them"""
    def filter(self, s):
        if not proxy3_options.options.get('remove-plugins', 0): return s
        s = remove_pattern(s, '<EMBED', '>')
        s = remove_pattern(s, '<embed', '>')
        s = remove_pattern(s, '<OBJECT', '>')
        s = remove_pattern(s, '<object', '>')
        return s

class SlashPreFilter(BufferSomeFilter): # Before generic ad removal
    BUFFER_LEN = 16

    def filter(self, s):
        if not proxy3_options.options.get('remove-slashdot-ads', 0): return s
        # Buggy HTML keeps the ad filter from working
        s = replace(s, '</ilayer>', '</layer>')
        s = replace(s, '</ILAYER>', '</layer>')
        return s

class SlashPostFilter(BufferSomeFilter): # After generic ad removal
    BUFFER_LEN = 256

    def filter(self, s):
        if not proxy3_options.options.get('remove-slashdot-ads', 0): return s
        # Remove the empty ad at the top
        s = replace(s, '<ILAYER height=90> </ILAYER>', '')
        s = replace(s, '<ILAYER height=90> </layer>', '') # PreFilter got it
        # Change the colors!!  How inconsistent their HTML is ..
        s = replace(s, 'bgcolor="#006666"', 'bgcolor="#000066"')
        s = replace(s, 'link="#006666"',    'link="#000066"')
        s = replace(s, 'color="#006666"',   'color="#000066"')
        s = replace(s, 'color=#006666',     'color="#000066"')
        s = replace(s, 'bgcolor="#ffffff"', 'bgcolor="#e1e1e1"')
        s = replace(s, 'bgcolor="#FFFFFF"', 'bgcolor="#e1e1e1"')
        s = replace(s, 'bgcolor=ffffff',    'bgcolor="#e1e1e1"')
        s = replace(s, 'bgcolor="ffffff"',  'bgcolor="#e1e1e1"')
        s = replace(s, 'bgcolor="#CCCCCC"', 'bgcolor=white')
        s = replace(s, 'bgcolor=cccccc',    'bgcolor=white')
        # Replace the custom message!
        i = find(s, 'This page was generated by ')
        if i >= 0:
            j = find(s, 'for', i)
            if j >= 0:
                s = (s[:i]+'This page was generated by'+
                     s[i:j]+" and then altered by an evil proxy"+s[j:])
        return s

class ExciteFilter(BufferAllFilter):
    def skip_to_table(self, s):
        i = find(lower(s), '<table')
        if i >= 0:  return s[:i], s[i:]
        else:       return s, ''

    def process_table(self, s, result, keep):
        # Assume we're at the <table ...>
        i = find(s, '>', 1)
        s0, s = s[:i], s[i:] # skip over the <table ...> tag
        if keep: result.append(s0)
        # Now look for another table or a close table
        while 1:
            sl = lower(s)
            i = find(sl, '<table')
            j = find(sl, '</table>')
            if i >= 0 and i < j:
                # A nested table
                if keep: result.append(s[:i])
                s = self.process_table(s[i:], result, keep)
            else:
                s0, s = s[:j+len('</table>')], s[j+len('</table>'):]
                if keep: result.append(s0)
                break
        return s
    
    def filter(self, s):
        # s = replace(s, 'border=0', 'border=2') # DEBUGGING TABLES
        N = 3
        result = []
        for i in range(N):
            s0, s = self.skip_to_table(s)
            result.append(s0)
            s = self.process_table(s, result, 0)
        result.append(s)
        return join(result)
    # We should also filter out the META REFRESH
    
class NewsTrackerFilter(BufferSomeFilter):
    BUFFER_LEN = 8192
    
    def filter(self, s):
        if not proxy3_options.options.get('remove-newstracker-ads', 0):return s
        s = remove_pattern(s, '<!-- Ad info:', '</td></tr></table><br>')
        s = remove_pattern(s, '<map name="news">', '</map>')
        s = remove_pattern(s, '<table width=120', '</table>')
        s = remove_pattern(s,
                           '<table border=0 cellpadding=0 '
                           'cellspacing=0 width=600>',
                           '</table>')
        s = replace(s, 'Click Here!', '')
        s = replace(s, '=99CCCC', '=F0F8F4')
        s = replace(s, '=FFDEAD', '=CCFFEE')
        s = replace(s, '<font face=arial size=-1>'
                    '<a href=http://www.excite.com/>Excite Home</a> '
                    '| <a href="/">News Home</a></font>',
                    '')
        s = replace(s, '<TABLE ALIGN=right', '<TABLE')
        s = replace(s,
                    '<img src="/img/news.gif" width=300 height=53 '
                    'border=0 usemap="#news">',
                    '<H1>Proxied News</H1>')
        return s
    
class MicrosoftCharacterFilter(BufferSomeFilter):
    BUFFER_LEN = 0
    
    TRANS=maketrans('\221\222\223\224', '`\'""')
    def filter(self, s):
        if not proxy3_options.options.get('fix-microsoft-charset', 0): return s
        return translate(s, self.TRANS)

class PopupAdFilter(BufferSomeFilter):
    BUFFER_LEN = 1024

    BodyPat = re.compile(r'(?is)<body\b.*?>')
    UnloadPat = re.compile(r'(?i)\bonunload\s*=\s*"?\'?([^\s">\']*)')
    
    WindowPat = re.compile(r'(?is)window.open\(["\'](.*?)["\'].*?\);')

    def filter_popups(self, s):
        import proxy3_block
        s = replace(s, 'TripodShowPopup();', '') # Remove Tripod popups ..
        results = []
        site = self.args['url']
        while 1:
            m = self.WindowPat.search(s)
            if m == None: break

            url = m.group(1) 
            url = urlparse.urljoin('http://'+site, url) 
            url = url[7:] 
            if proxy3_block.blocker.matches_blocked(url):
                # It's an ad .. eliminate!
                results.append(s[:m.start()])
                broadcast_event('ad-removed', url, 'window.open')
            else:
                # It's okay ..
                results.append(s[:m.end()])
            s = s[m.end():]
        results.append(s)
        return join(results, '')
    
    def filter(self, s):
        if not proxy3_options.options.get('remove-popups', 0): return s
        s = self.filter_popups(s)
        
        m = self.BodyPat.search(s)
        if m == None: return s

        tag = lower(m.group(0))
        p = self.UnloadPat.search(tag)
        if p == None: return s

        # It has BODY and UNLOAD, evil trick to keep you from closing a window
        return join([s[:m.start()+p.start()], s[m.start()+p.end():]], '')

class GenericAdFilter(BufferSomeFilter):
    BUFFER_LEN = 1024

    TagPat = re.compile(
            r'(?is)<(a|img|iframe|layer|ilayer|script|input|form'
            r'|applet|object|area|input|embed)\b.*?>')
    CommentStart = '<!--'
    CommentEnd   = '-->'
    
    # The i flag is case-insensitivity; s is to include newline in '.' 
    HrefPat = re.compile(r'(?i)href\s*=\s*"?\'?([^\s">\']*)')
    FormPat = re.compile(r'(?i)action\s*=\s*"?\'?([^\s">\']*)')
    SrcPat = re.compile(r'(?i)src\s*=\s*"?\'?([^\s">\']*)')
    AppletPat = re.compile(r'(?i)codebase\s*=\s*"?\'?([^\s">\']*)')

    in_comment = 0 # Class global, but reset to be object local
    
    def filter(self, s):
        if not proxy3_options.options.get('remove-ads', 0): return s
        
        import proxy3_block
        site = self.args['url']
        results = []
        while 1:
            if self.in_comment:
                # If we're in a comment, then we don't do ad filtering;
                # instead we only look for the end of the comment
                i = find(s, self.CommentEnd)
                if i < 0: break
                results.append(s[:i+len(self.CommentEnd)])
                s = s[i+len(self.CommentEnd):]
                self.in_comment = 0
                continue

            # We're not in a comment...
            m = self.TagPat.search(s)
            # In addition to the TagPat, we should look for CommentStart
            i = find(s, self.CommentStart)
            if i >= 0 and (m == None or i < m.start()):
                # Comment started or ended!
                results.append(s[:i+len(self.CommentStart)])
                s = s[i+len(self.CommentStart):]
                self.in_comment = 1
                continue
            
            if m == None: break

            # The tag matters because img comes alone, but others have </tag>
            tag = lower(m.group(1))
            if tag=='a':       srcpat = self.HrefPat
            elif tag=='form':  srcpat = self.FormPat
            elif tag=='applet':srcpat = self.AppletPat
            else:              srcpat = self.SrcPat

            # Save the first part of the string
            results.append(s[:m.start()])

            # Remove the section that was matched, if it's a blocked site
            link, s = s[m.start():m.end()], s[m.end():]
            # Now that we canonicalized the link, see if it's an ad
            p = None
            if proxy3_options.options.get('rewrite-urls', 0):
                p = srcpat.search(link)
            if p != None:
                url = p.group(1) # extract the url from the text
                url = urlparse.urljoin('http://'+site, url) # resolve url

                # First see if the original URL is something to remove
                if url[:7] == 'http://': url = url[7:]
                should_remove = proxy3_block.blocker.matches_blocked(url)

                if not should_remove:
                    # Attempt to rewrite the URL and check again
                    newurl = proxy3_rewrite.rewrite_url(url)
                    # Replace this portion of the text with the altered URL
                    if newurl and newurl != url:
                        # We don't do this all the time because we've
                        # done some alterations already and don't want
                        # them to affect the page .. only the */fwd/* stuff
                        if newurl[:1] == '/':
                            url = urlparse.urljoin(url, newurl)
                        else:
                            url = 'http://' + newurl
                        link = link[:p.start(1)]+url+link[p.end(1):]
                        if url[:7] == 'http://': url = url[7:]
                        if proxy3_block.blocker.matches_blocked(url):
                            should_remove = 1

                if should_remove:
                    if tag == 'img' or tag == 'embed':
                        # There's no end tag
                        link = ""
                    else:
                        # If it's not an img, remove until the matching </tag>
                        antitag = '</%s>' % tag
                        j = find(lower(s), antitag)
                        if tag == 'a':
                            # This is an inline element, so it can't
                            # contain block elements (p, td, etc.)
                            for endblock in ('</table>', '</tr>', '</td>',
                                             '</div>', '</p>'):
                                j1 = find(lower(s), endblock)
                                if 0 <= j1 < j:
                                    # The block ended earlier than the
                                    # inline tag, so there's an implicit
                                    # end tag at the end of the block ..
                                    j = j1
                                    antitag = endblock
                                    
                        if j >= 0: 
                            link = ""
                            s = s[j+len(antitag):]
                        else:
                            # We couldn't find the matching end tag,
                            # so make the filter window bigger for
                            # this document, and hope no tags are
                            # separated by more than 8k
                            self.BUFFER_LEN = 8192
                    
                    if link=='':
                        # We did filter stuff out
                        broadcast_event('ad-removed', url, tag)

            # Save the link
            results.append(link)

        # Return both (all the pieces we filtered) and (what we didn't)
        return join(results,''), s

from proxy3_util import *

# These filters are performed in order
register_filter('news.excite.com/news/*', 'text/html', NewsTrackerFilter)
#register_filter('www.excite.com/','text/html', ExciteFilter)
register_filter('*slashdot.org/*','text/html', SlashPreFilter)
register_filter('*','text/html', GenericAdFilter)
register_filter('*slashdot.org/*','text/html', SlashPostFilter)
register_filter('*','text/html', PopupAdFilter)
register_filter('*','text/html', MicrosoftCharacterFilter)
register_filter('*', 'text/html', NoScript)
register_filter('*', 'text/html', NoPlugins)
register_filter('lwn.net/*', '*', None) # It's nasty nasty burstnet code
register_filter('lwn.net/*', '*', NoScriptForce) # It's nasty nasty burstnet
