# Changes made by Insik Shin, ishin@cs.stanford.edu
#  Tue  2 Sep 1997 - Added communication with the UI through a second
#                    listener (at port 3331).  The second listener is
#                    in the UIListener class.  When the UI listener
#                    receives a request, it uses the UIPRcvRequest and
#                    UIPSndRequest classes to receive and send messages.
#                  - Changed HTTPRequest to keep track of the remote
#                    socket address.
#  Mon  8 Sep 1997 - Added Excite Ad filtering (with a small bug that
#                    only shows up if the ad was not received in one
#                    chunk)
#                  - Added control of ad filtering, site blocking, and
#                    Java class substitution into the UI
#                  - Added support in the proxy to control these features
#                  - Added UI support for querying the list of blocked sites

import socket
from string import *
from proxy3_base import Handler

UIPORT = 3331

###### User Interface classes

class UIPSndRequest(Handler):
    """User Interface Protocol Handler for sending UI messages."""

    def __init__ (self, host, port, message):
        self.host = host
	self.port = port
	self.message = message
	
    def process (self):
	# Create a new connection
	try:
	    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
	    s.connect((self.host, self.port))
	    s.setblocking(0)
	except socket.error, err:
	    # Error connecting, so don't return this handler
	    s.close()
	    # Don't print 111 because it just means the UI applet
	    # isn't up and running.
	    if err[0] != 111: print '*  ***  UIP Socket Error',err[0],err[1]
	    return []

	# Treat socket s like a pair of files (infile,outfile)
	self.infile = s.makefile('r')
	self.outfile = s.makefile('w',0)
	s.close()
	# send message
	self.outfile.write(self.message)

	return []


class UIPRequest(Handler):
    """User Interface Protocol Handler for receiving messages from UI.
       Accept UI's messages, and send their reply back to UI"""

    # Request number, to keep track of them all
    next_reqno = 1
    
    def __init__(self, infile, outfile, address):
	self.infile = infile
	self.outfile = outfile
	self.address = address
	# Assign this request a new number
	self.reqno = UIPRequest.next_reqno
	UIPRequest.next_reqno = UIPRequest.next_reqno+1

    def fileno(self):
	return self.infile.fileno()

    def __repr__(self):
	site = ' from %s' % self.address[0]
	if self.address[0] == '127.0.0.1':
	    site = ' (local)'
	return '<UIPRequest: %s%s>' % (self.reqno, site)
    
    def process(self):
	fi = self.infile
	fo = self.outfile

	line = fi.readline()
	while line[-1:] == '\r' or line[-1:] == '\n':
	    line = line[:-1]
	words = split(line)

	# Handle messages
	# print '       -->', words
	if words == ['CONNECT']:
	    fo.write("CONNECT OK\n")
	elif words == ['Ad','Filtering']:
	    options['remove-ads'] = 1	
	    fo.write("OK\n")
	elif words == ['No','Ad','Filtering']:
	    options['remove-ads'] = 0	
	    fo.write("OK\n")
	elif words == ['Class','Substitution']:
	    options['transform-classes'] = 1
	    fo.write("OK\n")
	elif words == ['No','Class','Substitution']:
	    options['transform-classes'] = 0
	    fo.write("OK\n")
	elif words == ['Site','Blocking']:
	    options['block-ads'] = 1	
	    fo.write("OK\n")
	elif words == ['No','Site','Blocking']:
	    options['block-ads'] = 0	
	    fo.write("OK\n")
	elif words == ['Asking','Blocked','Sites']:
	    for b in Blocker.blocked:
	        fo.write("     " + b +"\n");
	elif words == ['ALL']:
	    fo.write("MESSAGE ALL\n")
	elif words == ['MIDDLE']:
	    fo.write("MESSAGE MIDDLE\n")
	elif words == ['NOTHING']:
	    fo.write("MESSAGE NOTHING\n")
	else:
	    fo.write("UNKNOWN")	
		
	fi.close()
	fo.close()
	return []


class UIListener(Handler):
    """Accept and create connections from UI"""
    
    def __init__(self, UIPORT):	
	s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
	s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
	s.bind(('', UIPORT))
	s.setblocking(0)
	s.listen(1)
	self.conn = s

    def __repr__(self):
	return '<UIListener>'
    
    def fileno(self):
	return self.conn.fileno()

    def ping(self):
	# Listeners never die
	return [self]

    def process(self):
	try:
	    request, address = self.conn.accept()
	except socket.error:
	    # Connection attempted by UI, but aborted(?).
	    # Note that we must always include self in the return
	    # value for UIListener because the UIListener should always
	    # be kept up.
	    return [self]

	request.setblocking(1)
	handler = UIPRequest( request.makefile('r'),
			   request.makefile('w', 0),
			   address )

	request.close()

	# Return self AND the handler, indicating that both should
	# be put in the list of handlers
	return [self, handler]

#######  End of User Interface classes

import proxy3_options
proxy3_options.register_listener(UIListener, UIPORT)

def event_site_blocked(address, site, document):
    UIPSndRequest(address, UIPORT, "        -----> Blocked").process()
def event_process_request(obj, site, document, request, headers):
    UIPSndRequest(address, UIPORT, '* '+site+document).process()
    # Start the applet -- use server redirection
    if site=='_proxy' and document in ('/start/', '/start'):
        site = "http://%s:%s/http://_proxy/ui/" % (proxy3_base.HOST, PORT)
        html_output(obj.outfile, "", 302, ["Location: "+site]) 
        return []
    # Redirection to Insik's PUI page
    if site=='_proxy' and document[:4]=='/ui/':
        # All the pages on _proxy/ui/ have to be redirected
        site = 'www-cs-students.stanford.edu'
        if document=='/ui/': document = document + 'PUI.html'
        document = '/~ishin/proxy/' + document[4:]
        return site, document
    
