from string import *
from proxy3_util import *
from proxy3_event import broadcast_event
import socket, time
HOST = socket.gethostname()
TIMEOUT = 300.0 # How long a connection can be stale before we kill it

class Handler:
    """Handlers are objects that can be used as network connections
    (i.e., with a tofile method) but also have a method to process
    input from that connection."""

    priority = 50 # Integer value; higher priorities are handled first
    
    def __init__(self, parent):
	# If there is a parent, copy its fields into this object
	self.first_active = time.time()
	self.last_active = self.first_active
	if parent:
	    attr = vars(parent)
	    for a in attr.keys():
		setattr(self, a, attr[a])

    def is_readable(self):
        "Return 1 if this handler can be read from .. used in select()"
        return 1

    def is_writable(self):
        "Return 1 if this handler can be written to .. used in select()"
        return 0
    
    def ping(self):
	# This method should attempt to test the connection to see
	# if it is open, and return a list of new handlers
	if time.time() - self.last_active > TIMEOUT:
	    broadcast_event('kill-stale-handler', self)
            try:
                if self.socket: self.socket.close()
                if self.infile: self.infile.close()
                if self.outfile: self.outfile.close()
            except NameError:
                pass
	    return []
	else:
	    return [self]

class DataReader(Handler):
    """Read data from an input connection, and send it somewhere."""
    
    # This class serves both as a 'handler' (__repr__, process)
    # and as a 'receiver' (read, close)

    def __init__(self, parent, length):
	Handler.__init__(self, parent)
	self.bytes = 0
	self.length = length

    def __repr__(self, width=70):
	# __repr__ is a special function that is used to display the
	# object as a string
	if not self.length: # None or 0
	    progress = ' (%s bytes)' % self.bytes
	else:
	    progress = ' (%3s%% of %s)' % (100*self.bytes/self.length, 
					   self.length)
        name = shorten_url(self.site+self.document,width-11-len(progress))
        type = split(self.mimetype, '/')[-1]
        return '<%-5s<-- %s%s>' % (type[:5], name, progress)
    
    def ping(self):
	# Test the outbound connection by writing nothing
	try: self.outfile.write('')
	except IOError, err:
	    broadcast_event('io-error', self, err, '[ping]')
	    self.socket.shutdown(2)
            self.outfile.close()
            self.infile.close()
	    return []

	# Now try the inbound connection by reading nothing
	try: self.infile.read(0)
	except IOError, err:
	    if err.errno != 0:
		broadcast_event('io-error', self, err, '[ping]')
		self.socket.shutdown(2)
                self.outfile.close()
                self.infile.close()
		return []

	# Everything's okay
	return Handler.ping(self)

    def fileno(self):
	# Fileno is used to tell the system which file handle is being
	# waited on
	return self.infile.fileno()

    def process(self, retries=0):
	try: 
	    data = self.infile.read()
            if not data and self.length:
                print 'EOSTREAM?????????????', self.site, self.document, `self.bytes`, '/', `self.length`
	except IOError, err:
            if err.errno == 0:
                # If it's 0, it means end of stream, but note that
                # data=='' also means end of stream (!).
                data = None
            elif err.errno == 11:
                # If it's 11 (EAGAIN), it means we tried to read
                # when there wasn't any data available.  (!?!)
                if retries < 5:
                    time.sleep(1)
                    return self.process(retries+1)
		broadcast_event('unix-error')
                return [self]
            else:
                # Broken pipe?
                broadcast_event('io-error', self, err)
                data = None
	if not data:
	    # This connection is dead
	    self.close()
	    return []
	else:
	    # This connection may have more data
	    self.bytes = self.bytes + len(data)
	    self.last_active = time.time()
	    try: self.read(data)
	    except IOError, err:
		# Broken pipe?
		broadcast_event('io-error', self, err)
		if self.socket: self.socket.shutdown(2)
		self.close()
		return []
	    if self.length and self.bytes >= self.length: 
		self.close()
                broadcast_event('status', None, None, self)
		return [] # We've read enough
	    return [self]

    def write(self, data):
	"Oddly enough, write(x) is an alias for read(x)"
	return self.read(data)

    def read(self, data):
	raise "DataReader:  unimplemented method"

    def close(self):
	self.infile.close()


class CopyData(DataReader):
    """Directly send any data that is read to an output connection"""

    def __init__(self, parent, length):
	DataReader.__init__(self, parent, length)

    def read(self, data):
	self.outfile.write(data)

    def close(self):
	DataReader.close(self)
	self.outfile.close()

