import socket, time

# Let's see if we can get to the signal module
try:
    import signal
    try:
        signal.SIGALRM
        USE_ALARM = 20
    except AttributeError:
        USE_ALARM = 0
    # print 'Loading signal handlers...'
except ImportError:
    USE_ALARM = 0

class NameLookup:
    TIMEOUT_NORMAL = 1800.0 # thirty minutes
    TIMEOUT_ERR = 60.0 # one minute
    cache = {}

def lookup_name(site, use_alarm=USE_ALARM):
    if NameLookup.cache.has_key(site):
        expire_time, result = NameLookup.cache[site]
        if time.time() < expire_time:
            if type(result) == type(''):  return result
            else:                         raise socket.error, result

    # It's not in the cache, or it's too old
    try: 
        if use_alarm:
            def signal_handler(signum, frame):
                raise socket.error, 'Timeout'
            import signal
            signal.signal(signal.SIGALRM, signal_handler)
            signal.alarm(use_alarm)
        try: 
            result = socket.gethostbyname(site)
        finally: 
            if use_alarm: signal.alarm(0)
        NameLookup.cache[site] = (time.time()+NameLookup.TIMEOUT_NORMAL,
                                  result)
        return result
    except socket.error, err:
        if err != 'Timeout':
            NameLookup.cache[site] = (time.time()+NameLookup.TIMEOUT_ERR, err)
        raise socket.error, err
