# Basic HTTP stuff, including a listener
import proxy3_options
PORT = proxy3_options.options.get('port', 3333)

HTML_ERROR = '''
<HTML>
  <HEAD>
    <TITLE>Could not connect</TITLE>
    <STYLE TYPE="text/css"><!--
      SPAN.a { font-family: georgia, times; font-size: 36pt; }
      SPAN.b, P.retry { font-family: tahoma, verdana, arial; font-size:200%%; }
      P.retry { text-align: left; }
    --></STYLE>
  </HEAD>
  <BODY BGCOLOR=#800000 TEXT=white LINK=#ff8888 VLINK=#ff88aa>
    <TABLE WIDTH="100%%" HEIGHT="100%%" CELLPADDING=10 CELLSPACING=0>
      <TR>
        <TD ALIGN=center WIDTH=25%% BGCOLOR=#805050>
          <P><SPAN CLASS=a>Error:</SPAN></P>
        </TD>
        <TD BGCOLOR=#805050>
          <P><SPAN CLASS=b>Could not connect</SPAN></P>
        </TD>
      </TR>
      <TR>
        <TD ALIGN=right WIDTH=25%% BGCOLOR=#805050>
          <P>%(site)s<BR>(%(site_ip)s)</P>
        </TD>
        <TD BGCOLOR=white>
          <FONT COLOR=black>
            <P>While trying to retrieve %(document)s:</P>
            <PRE>
              %(err)s
            </PRE>
          </FONT>
          <P CLASS=retry><A HREF="http://%(site)s%(document)s">Retry?</A></P>
        </TD>
      </TR>
    </TABLE>
  </BODY>
</HTML>
'''
  
import socket, rfc822, time
from string import *

import proxy3_filter
import proxy3_base
from proxy3_dns import lookup_name
from proxy3_base import Handler
from proxy3_util import *
from proxy3_event import broadcast_event

class HTTPReply(Handler):
    def __init__(self, parent, infile):
        Handler.__init__(self, parent)
        self.infile = infile
        self.socket.setblocking(1)

    def __repr__(self, width=70):
        name = shorten_url(self.site+self.document,width-16)
        return '<HTTP <-- %s %s>' % (self.response_code[0], name)
    
    def fileno(self):
        return self.infile.fileno()

    def process(self):
        self.reply = self.infile.readline()
        # NOTE: I'm using blocking I/O here so if the headers don't all
        # arrive at once, we're in trouble.  This really should be reading
        # into a string, and then once we have two newlines, we should turn
        # the string into a StringIO object that gets passed into rfc822.
        # The reason it's not done that way is that we have to figure out
        # what to do with the leftover string that's past the end of headers..
        self.serverheaders = rfc822.Message(self.infile, 0)
        broadcast_event('process-reply', self, 
                        self.site, self.document, 
                        self.reply, self.serverheaders)

        self.last_active = time.time()

        # If there is a content-length, we should only read that amount
        length = self.serverheaders.getheader('Content-Length')
        if length: 
            try: length = atoi(length)
            except: length = None
        else: length = None

        # Should we try to reuse this connection later? (unimplemented)
        try: http_version = split(self.reply)[0]
        except IndexError: http_version = 'HTTP/0.9'

	# Get the HTTP status code
	no_body = ('204', '304')
        if self.reply:
            try: code = split(self.reply)[1]
            except IndexError: code = ''
	else: code = ''

        # HACK -- instead of 404, redirect to Google's cached copy
        if code == '404':
            self.outfile.write(self.reply)
            self.outfile.write(join(self.serverheaders.headers, '')+'\r\n')
            self.outfile.write('<html><head><style><!--\r\n'
                               'h1 { color: red; font: 72pt arial,verdana; }\r\n'
                               '// -->\r\n'
                               '</style><title>404</title></head><body><h1>404</h1> Try <a href="http://www.google.com/search?q=cache:%s%s">Google\'s cached copy...</a></body></html>' % (self.site, self.document))
            self.infile.close()
            self.outfile.close()
	    return []
            
        #if not length and code not in no_body: stay_open = 0
        #elif http_version == 'HTTP/1.1': # NOTE: really should be >= 1.1
        #    stay_open = 1 # Assume connection can stay open
        #    c = self.serverheaders.getheader('Connection')
        #    if c and lower(c) == 'close': stay_open = 0
	#    if not c: stay_open = 0 # DEBUG - don't use implicit keep-opens
        #elif http_version == 'HTTP/1.0':
        #    stay_open = 0
        #    c = self.serverheaders.getheader('Connection')
        #    if c and lower(c) == 'keep-alive': stay_open = 1
	#else: stay_open = 0

        # Some replies have headers but no bodies
	if code in no_body:
	    # 304: Not Modified, no data remains
	    # 204: No response, no data to read
            # stay_open = 0
            # Send the reply to the browser
            self.outfile.write(self.reply)
            self.outfile.write(join(self.serverheaders.headers, '')+'\r\n')
            self.infile.close()
            self.outfile.close()
	    return []

        # Find the filters that may have to be run, build a stream of them
        stream = self.create_stream(length)
        # Now rewrite the header if we need to
        if self.mimetype == 'text/html':
            # It's probably filtered, so remove the content length EW :( :(
            self.serverheaders.headers = header_rewrite(
                self.serverheaders.headers, 'Content-Length', '')
        self.outfile.write(self.reply)
        self.outfile.write(join(self.serverheaders.headers, '')+'\r\n')

        self.socket.setblocking(0)

	# There's probably data, so let's read it right away
	return stream.process()

    def create_stream(self, length):
        try: self.mimetype = self.serverheaders['Content-type']
        except KeyError: self.mimetype = 'text/plain'

        # Adjust this object's priority based on the MIME type...
        # this should go elsewhere
        if find(self.mimetype, 'image/') >= 0:
            # Images get low priorities
            self.priority = 10
        elif find(self.mimetype, 'text/') >= 0:
            # Text gets a high priority
            self.priority = 75
            
        stream = proxy3_filter.get_filter(self.original_request,
                                          self.mimetype, self, length,
                                          self.clientheaders,
                                          self.serverheaders)

        # If we have chunked encoding, add the unchunking filter
        if self.serverheaders.getheader('Transfer-Encoding') == 'chunked':
            eh = 0
            if self.serverheaders.getheader('Trailer'): eh = 1
            stream = proxy3_filter.UnChunkingFilter(self, stream, length,
                                                    extra_headers = eh)
        return stream

def transform_headers(headers):
    headers = header_rewrite(headers, 'Proxy-Connection', '')
    if not proxy3_options.options.get('rewrite-headers', 1): return headers
    headers = header_rewrite(headers, 'Referer', 
                             proxy3_options.options.get('referer',lambda x:x))
    headers = header_rewrite(headers, 'User-Agent',
                             proxy3_options.options.get('user-agent',lambda u:u))
    #headers = header_rewrite(headers, 'Connection', 'keep-alive')
    headers = header_rewrite(headers, 'Connection', 'close')
    headers = header_rewrite(headers, 'Accept-Encoding', 'identity')
    return headers

class HTTPRequest(Handler):
    def __init__(self, parent, socket, req, site, document, clientheaders):
        Handler.__init__(self, parent)
        self.socket = socket
        self.request = req
        self.site = site
        self.document = document
        self.clientheaders = clientheaders
        
    def fileno(self):
        return self.socket.fileno()

    def is_readable(self):
        return 0

    def is_writable(self):
        return 1
    
    def __repr__(self, width=70):
        name = shorten_url(self.site+self.document,width-16)
        return '<HTTP --> %s %s>' % (self.response_code[0], name)
    
    def process(self):
        # Now that the socket is writable, send out a request
        try:
            self.socket.send(self.request)
            self.socket.send('\r\n')
        except socket.error, err:
            site = self.site
            if self.port != 80: site = '%s:%d' % (site, self.port)
            document = self.document
            try: site_ip = lookup_name(site)
            except: site_ip = site
            html_output(self.outfile, HTML_ERROR % vars(), code=500)
            broadcast_event('socket-error', self, err)

            self.infile.close()
            self.socket.close()
            return []
        
        # Treat input from socket s2 like a file (in2)
        in2 = self.socket.makefile('rb')
        
        # Forward the request to the remote server
        self.socket.send(join(self.clientheaders.headers, '')+'\r\n')
        # Additional data may be sent along with this request
        if 'content-length' in self.clientheaders.keys():
            # There is additional data to transfer, add 1 to get newline
            # Note: the add 1 caused hangs in some forms, so I took it out
            data = self.infile.read(atoi(self.clientheaders['content-length']))
            broadcast_event('extra-form-data', self, data)
            self.socket.send(data)
            self.socket.send('\r\n') # Send this - fixes some problems
            
        # Shut down one side of the socket - no more writes needed
        # This doesn't work with some servers (like Medusa)
        #self.socket.shutdown(1)

        # Handler for the reply:
        # This handler will copy the reply headers from the web server
        # to the browser, then invoke the chained handler.  Note that we 
        # kill the current handler by not including it.
        handlers = []
        if self.response_code[0] != 'HEAD':
            handlers = [ HTTPReply(self, in2) ]

        # Assume we need no more data from the connection 
        # If we don't close this, then we'll end up with a confused browser
        self.infile.close()
        
        return handlers

class HTTPConnect(Handler):
    def __init__(self, parent, response_code):
        Handler.__init__(self, parent)
        self.response_code = response_code

    def fileno(self):
        return self.infile.fileno()

    def __repr__(self, width=70):
        return '<HTTP --> %s %s>' % (self.response_code[0], self.reqno)
    
    def process(self):
        fi = self.infile
        fo = self.outfile

        clientheaders = rfc822.Message(fi, 0)
        self.last_active = time.time()

        # This allows us to use the proxy without setting up a proxy
        # in the web browser.  Just use a URL of the form:
        #     http://proxymachine:3333/http://...
        if self.response_code[1][:8] == '/http://':
            self.response_code[1] = self.response_code[1][1:] # Strip off /

        # Figure out where we want to connect
        site, document = stripsite(self.response_code[1])
        if not site and 'Host' in clientheaders.keys():
            site = clientheaders['Host']
        # If there was no machine name, then consider it a special URL
        if not site: site = '_proxy'

        # Let proxy:* URLs map to http://_proxy/*
        if site[:6] == 'proxy:':
            site, document = '_proxy', site[6:]
            # Hack the slashes a bit .. :(
            if document[:1] != '/': document = '/'+document
            if document[-1] == '/': document = document[:-1]

        # Before redirection, let's save what the user requested
        self.original_request = site+document
        self.referer = clientheaders.getheader('Referer')
        
        # Display the site/document pair
        out_headers = transform_headers(clientheaders.headers)
        out_headers = header_rewrite(out_headers, 'Host', site) # HTTP/1.1
        clientheaders.headers = out_headers # DANGEROUS?!
        result = broadcast_event('process-request', self,
                                 site, document,
                                 join(self.response_code, ' '), clientheaders)

        if type(result) == type(()):
            # It's a pair -- hopefully a new site, document
            site, document = result
        if type(result) == type([]):
            # It returned a list, which means some handler took care of this
            if self not in result:
                # We're being closed!
                fi.close()
                fo.close()
            return result

        # If the site is _proxy then we intercept
        if site == '_proxy':
            # Couldn't find a handler ..
            html_output(fo, "The proxy does not have a handler for %s" %
                        document, code=404)
            fi.close()
            fo.close()
            return []
        
        # If the site has a : in it, that means that it's
        # a site followed by a port number
        self.port = 80
        i = find(site,':')
        if i >= 0:
            try:
                self.port = atoi(site[i+1:])
                site = site[:i]
            except ValueError:
                # It wasn't a number, so maybe it's not really a port
                i = find(site[:i], '.')
                if i < 0:
                    # It looks like it may be a non-http:// URL
                    html_output(fo, "No support for non-http yet.", 404)
                    fi.close()
                    return []

        # Find out if there's a parent proxy .. if there is, we don't
        # need to perform name lookup.  Whee.
        parent_proxy = proxy3_options.options.get('proxy', None)
        if type(parent_proxy) == type('') and strip(parent_proxy):
            parent_proxy = split(parent_proxy, ':')
            parent_proxy = (parent_proxy[0], int(parent_proxy[1]))
        if type(parent_proxy) != type(()):
            parent_proxy = None

        if parent_proxy:
            site_ip = site
        else:
            name_lookup_time = time.time()
            try:
                broadcast_event('name-lookup-begin', site)
                site_ip = lookup_name(site)
                broadcast_event('name-lookup-end', site, site_ip)
            except socket.error, err:
                # If there were no dots in the hostname,
                # let's try again, with WWW.*.COM
                if find(site, '.') < 0:
                    # Server redirection
                    html_output(fo, "", 302, [
                        "Location: http://www.%s.com/" % site]) 
                    print '|  -->  Redirecting %s to http://www.%s.com/' % (
                        site, site)
                    fi.close()
                    return []
                # Otherwise, let's hope that the next attempt works
                site_ip = site
            name_lookup_time = time.time() - name_lookup_time
            broadcast_event('time-name-lookup', site, document,
                            name_lookup_time)

        try:
            # Get a connection to the site (asynchronously)
            s2 = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            #s2.setsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF, 131072)
            #s2.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, 131072);
            s2.setblocking(0)
            if parent_proxy:
                # If we're going through a parent proxy,
                # the request is different, and we connect to
                # a different site
                port_str = ''
                if self.port != 80: port_str = ':%s' % self.port
                req = '%s http://%s%s%s %s' % (self.response_code[0],
                                               site, port_str, document,
                                             self.response_code[2])
                s2.connect_ex((parent_proxy[0], parent_proxy[1]))
            else:
                s2.connect_ex((site_ip, self.port))
                req = '%s %s %s' % (self.response_code[0], document, 
                                    self.response_code[2])
        except socket.error, err:
            # Error connecting, so don't return this handler
            html_output(fo, HTML_ERROR % vars(), code=500)
            broadcast_event('socket-error', self, err)

            fi.close()
            return []

        return [HTTPRequest(self, s2, req, site, document, clientheaders)]
    
class Request(Handler):
    """Accept a request from the web browser, and send it to the web
    server."""

    # Request number, to keep track of them all
    next_reqno = 1
    
    def __init__(self, parent, infile, outfile, address):
        Handler.__init__(self, parent)
        self.infile = infile
        self.outfile = outfile
        self.address = address
        # Assign this request a new number
        self.reqno = Request.next_reqno
        Request.next_reqno = Request.next_reqno+1

    def fileno(self):
        return self.infile.fileno()

    def __repr__(self, width=70):
        site = 'from %s' % self.address[0]
        if self.address[0] == '127.0.0.1':
            site = '(local)'
        return '<TCP %-4d %s>' % (self.reqno, site)

    def process(self):
        fi = self.infile
        fo = self.outfile

        line = fi.readline()
        while line[-1:] == '\r' or line[-1:] == '\n':
            line = line[:-1]
        words = split(line)
        self.last_active = time.time()

        if len(words) == 2: words.append('HTTP/0.9')
        if len(words) != 3 or not (words[0] in ['HEAD','GET','POST']):
            # Send error 400
            html_output(fo, "<BODY>Bad Request: %s</BODY>" % line, code=400)
            broadcast_event('unknown-request', self, line)

            # Nothing left to do here, so don't return self
            fi.close()
            return []
        else:
            # Remove CR from the URL
            words[1] = replace(words[1], '\r', '')
            return HTTPConnect(self, words).process()


class Listener(Handler):
    """Accept and create connections from web browsers"""

    priority = 100 # High priority for listeners
    
    def __init__(self, PORT):   
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        interface = ''
        if proxy3_options.options.get('localonly', 0): interface = 'localhost'
        s.bind(('', PORT))
        s.setblocking(0)
        s.listen(5)
        self.conn = s
        self.port = PORT

    def __repr__(self, width=70):
        return '<Listener HTTP:%d>' % PORT

    def ping(self):
        # Listeners should never be killed
        return [self]

    def fileno(self):
        return self.conn.fileno()
    
    def process(self):
        self.last_active = time.time()
        try:
            request, address = self.conn.accept()
        except socket.error:
            # Connection attempted by browser, but aborted(?).
            # Note that we must always include self in the return
            # value for Listener because the Listener should always
            # be kept up.
            return [self]

        # Create a handler for this connection
        # Note that this connection has to be blocking, unfortunately,
        # because after it is select()-ed, the RFC822 header library
        # needs to read all the headers at once.  If the connection is
        # non-blocking, then the headers might be read only partially,
        # and then HTTPRequest will be confused.

        #request.setsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF, 131072)
        #request.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, 131072);
        request.setblocking(1)
        try:
            handler = Request( None, 
                               request.makefile('rb'),
                               request.makefile('wb', 0),
                               address )
        except:
            # We couldn't make the handler, so ignore the connection
            return [self]
        
        # Note that Python is garbage collected, and although this
        # particular object is being closed, the socket itself won't
        # be closed until all references (in particular, the two
        # makefiles above) are closed.
        request.close()

        # Return self AND the handler, indicating that both should
        # be put in the list of handlers
        return [self, handler]

import proxy3_options, proxy3_block
proxy3_options.register_listener(Listener, PORT)
